
#include <unistd.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "cfgfile.h"
#include "config.h"
#include "log.h"
#include "conversd.h"
#include "access.h"
#include "hmalloc.h"

char *myhostname = NULL;
char *mysysinfo = NULL;
char *myemailaddr = NULL;
char *mysyslink = NULL;
char *mytimezone = NULL;
int listen_port = 3600;
int secretnumber = 0;
int defaultchannel = 0;
int restrictedmode = 0;
int max_u_queue = 50 * 1024;
int max_h_queue = 100 * 1024;
int demonize = 1;
int tryonly = 0;

char conffile[PATH_MAX];

int do_security(int *dest, int argc, char **argv);
int do_link(struct permlink ***pa, int argc, char **argv);
int do_maxlinks(struct permlink ***pa, int argc, char **argv);

/*
 *	Configuration file commands
 */

#define _CFUNC_ (int (*)(void *dest, int argc, char **argv))

static struct cfgcmd cfg_cmds[] = {
	{ "access",		_CFUNC_ do_access,	NULL		},
	{ "defaultchan",	_CFUNC_ do_int,		&defaultchannel	},
	{ "email",		_CFUNC_ do_string,	&myemailaddr	},
	{ "link",		_CFUNC_	do_link,	&permarray	},
	{ "log",		_CFUNC_ do_log,		NULL		},
	{ "logdest",		_CFUNC_ do_logdest,	NULL		},
	{ "maxhqueue",		_CFUNC_ do_int,		&max_h_queue	},
	{ "maxuqueue",		_CFUNC_ do_int,		&max_u_queue	},
	{ "maxlinks",		_CFUNC_ do_maxlinks,	&permarray	},
	{ "port",		_CFUNC_ do_int,		&listen_port	},
	{ "secretnum",		_CFUNC_ do_int,		&secretnumber	},
	{ "security",		_CFUNC_	do_security,	&restrictedmode },
	{ "server",		_CFUNC_ do_string,	&myhostname	},
	{ "sysinfo",		_CFUNC_ do_string,	&mysysinfo	},
	{ "syslink",		_CFUNC_ do_string,	&mysyslink	},
	{ "syslog",		_CFUNC_ do_syslog,	NULL		},
	{ "timezone",		_CFUNC_	do_string,	&mytimezone	},
	{ NULL,			NULL,			NULL		}
};

/*
 *	Configure a bitmask
 */
 
int bitcfg(struct bitlist_t *bitlist, int argc, char **argv)
{
	struct bitlist_t *bl;
	int i;
	int mask = 0;
	
	for (i = 0; i < argc; i++) {
		bl = bitlist;
		while (bl->name) {
			if (!strcasecmp(argv[i], bl->name)) {
				mask |= bl->value;
				break;
			}
			bl++;
		}
		if (bl->name == NULL) {
			log(L_CRIT, "Unknown parameter: \"%s\"", argv[i]);
			return -1;
		}
	}
	
	return mask;
}

/*
 *	Print a list of enabled bits to a buffer
 */

int printbits(int mask, struct bitlist_t *bitlist, char *buf, size_t len)
{
	struct bitlist_t *bl;
	
	bl = bitlist;
	while (bl->name) {
		if (mask & bl->value) {
			strncat(buf, " ", len);
			strncat(buf, bl->name, len);
		}
		bl++;
	}
	
	buf[len-1] = '\0';
	
	return strlen(buf);
}

/*
 *	Configure an integer
 */
 
int numcfg(struct bitlist_t *bitlist, char *arg)
{
	struct bitlist_t *bl;
	int val = 0;
	
	bl = bitlist;
	while (bl->name) {
		if (!strcasecmp(arg, bl->name)) {
			val = bl->value;
			break;
		}
		bl++;
	}
	if (bl->name == NULL) {
		log(L_CRIT, "Unknown parameter: \"%s\"", arg);
		return -1;
	}
	
	return val;
}

/*
 *	Print the name of a number to a buffer
 */

int printnum(int val, struct bitlist_t *bitlist, char *buf, size_t len)
{
	struct bitlist_t *bl;
	
	bl = bitlist;
	while (bl->name) {
		if (val == bl->value) {
			strncat(buf, " ", len);
			strncat(buf, bl->name, len);
		}
		bl++;
	}
	
	buf[len-1] = '\0';
	
	return strlen(buf);
}

/*
 *	do_security
 */

int do_security(int *dest, int argc, char **argv)
{
	if (argc < 2)
		return -1;
		
	if (*argv[1] == 'r') {
		*dest = 1;
		log(L_CONFIG, "Security: Restricted");
	} else if (*argv[1] == 'o') {
		*dest = 2;
		log(L_CONFIG, "Security: Observer");
	} else {
		log(L_ERR, "Unknown Security mode: %s", argv[1]);
		return -1;
	}
	
	return 0;
}

/*
 *	do_link
 */

int do_link(struct permlink ***pa, int argc, char **argv)
{
	char *host_name;
	char *sock_name;
	char *primary_link;
	char *cp;
	struct permlink *p, *p2;
	
	if (argc < 3 || argc > 4)
		return -1;
	
	host_name = argv[1];
	sock_name = argv[2];
	primary_link = argv[3];
	
	p = update_permlinks(host_name, NULLCONNECTION, 1);
	if (!p)
		return -1;
		
	strncpy(p->name, host_name, HOSTNAMESIZE);
	if ((cp = strchr(sock_name, ':'))) {
		*cp = '\0';
		cp++;
		p->port = atoi(cp);
	} else
		p->port = 3600;
		
	p->host = hstrdup(sock_name);
	p->group = 1;
	p->quality = 1;
	p->primary = NULL;
	/* p->backup = NULL; */
	
	/* If we have a primary link, find it and link it in */
	if (primary_link) {
		p2 = find_permlink(primary_link);
		if (p2) {
			p->waittime = 30;
			p->primary = p;
			/* p2->backup = p; */
		} else {
			log(L_ERR, "Link: %s: wants to be a backup for %s, which has not been defined yet! Discarding this link.",
				host_name, primary_link);
			
		}
	}
	
	log(L_CONFIG, "Link: %s [%s]%s%s%s", p->name, p->host,
		(p->primary) ? " (backup to " : "",
		(p->primary) ? p->primary->name : "",
		(p->primary) ? ")" : "");
	
	return 0;
}

/*
 *	do_maxlinks
 */

int do_maxlinks(struct permlink ***pa, int argc, char **argv)
{
	struct permlink **tmp;
	int links = NR_PERMLINKS;
	
	if (argc < 2)
		return -1;
	
	NR_PERMLINKS = atoi(argv[1]);
	tmp = permarray;
	permarray = hcalloc(NR_PERMLINKS, sizeof(struct permlink *));
	if (tmp != NULLPERMARRAY) {
		memcpy(permarray, tmp, links * sizeof(struct permlink *));
		hfree(tmp);
	}
	
	log(L_CONFIG, "MaxLinks: %d", NR_PERMLINKS);
	
	return 0;
}

/*
 *	Parse command line arguments
 */

void parse_cmdl(int argc, char **argv)
{
	int c;
	
	while ((c = getopt(argc, argv, "ft")) != -1) {
	switch (c) {
		case 'f':
			demonize = 0;
			break;
		case 't':
			tryonly = 1;
			break;
		default:
			fprintf(stderr, "Ugh: Unknown parameter '%c'. Known: -f (don't fork), -t (try only)\n", c);
			exit(1);
	}
	}
	
}

/*
 *	Read configuration files
 */

void read_config(void)
{
	defaultchannel = 0;
	secretnumber = 0;
	restrictedmode = 0;
	
	if (mysysinfo) {
		hfree(mysysinfo);
		mysysinfo = NULL;
	}
	if (myemailaddr) {
		hfree(mysysinfo);
		mysysinfo = NULL;
	}
	if (myhostname) {
		hfree(myhostname);
		myhostname = NULL;
	}
	
	if (read_cfgfile(conffile, cfg_cmds))
		exit(1);
	
	if (!( (myhostname) && (myemailaddr) && (mysysinfo) && (listen_port)
	   && (secretnumber) )) {
	   	log(L_CRIT, "Missing entries in configuration file!");
	   	exit(1);
	}
	
	if (secretnumber >= INT_MAX) {
		log(L_CRIT, "Secret number is larger than MAX_INT (%d)!", INT_MAX);
		exit(1);
	}
	
	if ((defaultchannel < 1) || (defaultchannel > 32767)) {
		log(L_CRIT, "Default channel is smaller than 1 or larger than 32767.");
		exit(1);
	}
	
	if (strlen(myhostname) > 9) {
		log(L_CRIT, "Server hostname is longer than 9 characters!");
		exit(1);
	}
	
	if ((strchr(myhostname, ' ')) || (strchr(myhostname, '\t'))) {
		log(L_CRIT, "Server hostname contains whitespace!");
		exit(1);
	}
}

